import gzip
from typing import List, Dict, Any

import pandas as pd  # type: ignore
from csverve.errors import CsverveInputError


class IrregularCsverveInput(object):
    def __init__(self, filepath: str, dtypes: Dict[str, str], sep=',') -> None:
        """
        CSV file and all related metadata.

        @param filepath: Path of CSV.
        @param dtypes: dictionary of pandas dtypes by column, keys = column name, value = dtype.
        """
        self.filepath: str = filepath

        self.sep = sep
        self.columns = self.get_columns()

        self.dtypes: Dict[str, str] = dtypes

    @property
    def __file_type(self) -> str:
        if self.filepath.endswith('gz'):
            return 'gzip'
        elif self.filepath.endswith('csv'):
            return 'plain-text'
        else:
            raise CsverveInputError('Unsupported file type: {}'.format(self.filepath))

    @property
    def yaml_file(self) -> str:
        """
        Append '.yaml' to CSV path.

        @return: YAML metadata path.
        """
        return self.filepath + '.yaml'

    def get_columns(self) -> List[str]:
        """
        Detect whether file is tab or comma separated from header.
        @return: '\t', or ',', or raise error if unable to detect separator.
        """
        opener: Any = gzip.open if self.__file_type == 'gzip' else open
        with opener(self.filepath, 'rt') as inputfile:
            header: str = inputfile.readline().strip()

        columns: List[str] = header.split(self.sep)

        return columns

    def read_csv(self, chunksize: int = None) -> pd.DataFrame:
        """
        Read CSV.

        @param chunksize: Number of rows to read at a time (optional, applies to large datasets).
        @return: pandas DataFrame.
        """
        try:
            data: pd.DataFrame = pd.read_csv(
                self.filepath, chunksize=chunksize,
                sep=self.sep, names=self.columns, dtype=self.dtypes
            )
        except pd.errors.EmptyDataError:
            data = pd.DataFrame(columns=self.columns)

        if chunksize:
            for df in data:
                for col in data.columns.values:
                    assert col in self.dtypes, col
                yield df
        else:
            for col in data.columns.values:
                assert col in self.dtypes, col
            return data



import os
from typing import List, Dict, Union, Any

import pandas as pd  # type: ignore
import yaml
from csverve.errors import CsverveParseError


class CsverveInput(object):
    def __init__(self, filepath: str) -> None:
        """
        CSV file and all related metadata.

        @param filepath: Path of CSV.
        """

        self.filepath: str = filepath

        self._verify_input()

        self._yamldata = self._load_yaml()

    @property
    def header(self) -> bool:
        """
        True if file has header

        @return: header
        """
        return self._yamldata['header']

    @property
    def separator(self) -> str:
        """
        get the separator used

        @return: separator
        """
        return self._yamldata['sep']

    @property
    def columns(self) -> List[str]:
        """
        get the list of columns

        @return: separator
        """
        return [val['name'] for val in self._yamldata['columns']]

    @property
    def dtypes(self) -> Dict[str, str]:
        """
        get the data types

        @return: dtypes
        """
        return {val['name']: val['dtype'] for val in self._yamldata['columns']}

    @property
    def yaml_file(self) -> str:
        """
        Append '.yaml' to CSV path.

        @return: YAML metadata path.
        """
        return self.filepath + '.yaml'

    def _load_yaml(self) -> Dict[str, Any]:
        """
        load the yaml data

        @return: Dict
        """
        with open(self.yaml_file, 'rt') as yamlfile:
            yamldata = yaml.safe_load(yamlfile)
        return yamldata

    def _verify_input(self):
        """
        Verify gzip status and check for yaml

        @return:
        """
        if not self.filepath.endswith('.gz'):
            raise CsverveParseError('input must be gzipped')

        if not os.path.exists(self.yaml_file):
            raise CsverveParseError('yaml file missing')

    def _cast_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Cast dataframe dtypes.

        @param df: Pandas DataFrame.
        @return: Pandas DataFrame.
        """
        for column_name in df.columns.values:
            df[column_name] = df[column_name].astype(self.dtypes[column_name])
        return df

    def _verify_data(self, df: pd.DataFrame, columns) -> None:
        """
        Verify columns of DataFrame match those of class property.

        @param df: Pandas DataFrame.
        @return:
        """
        if not set(list(df.columns.values)) == set(columns):
            raise CsverveParseError("metadata mismatch in {}".format(self.filepath))

    def read_csv(self, chunksize: int = None, usecols=None, dtype=None) -> pd.DataFrame:
        """
        Read CSV.

        @param chunksize: Number of rows to read at a time (optional, applies to large datasets).
        @param usecols: Restrict to specific columns (optional).
        @param dtype: Override the dtypes on specific columns (optional).
        @return: pandas DataFrame.
        """

        def return_gen(df_iterator, columns):
            for df in df_iterator:
                self._verify_data(df, columns)
                yield df

        # if header exists then use first line (0) as header
        header: Union[int, None] = 0 if self.header else None
        names: Union[None, List[str]] = None if self.header else self.columns
        columns: List[str] = usecols if usecols else self.columns

        # Override dtypes
        final_dtype = self.dtypes
        if dtype is not None:
            for name, dtype in dtype.items():
                if name in final_dtype:
                    final_dtype[name] = dtype
                else:
                    raise ValueError(f'dtype column {name} not present')

        try:
            data: pd.DataFrame = pd.read_csv(
                self.filepath,
                compression='gzip',
                chunksize=chunksize,
                sep=self.separator,
                header=header,
                names=names,
                dtype=final_dtype,
                usecols=usecols
            )
        except pd.errors.EmptyDataError:
            data = pd.DataFrame(columns=columns)
            data = self._cast_dataframe(data)

        if chunksize:
            return return_gen(data, columns)
        else:
            self._verify_data(data, columns)
            return data



#!/usr/bin/env python

"""Tests for `csverve` package."""

import itertools
import os
import random
import unittest

import csverve.api as api
import csverve.utils as utils
import numpy as np
import pandas as pd
import pytest
from click.testing import CliRunner
from csverve import cli
from csverve.core.csverve_input import CsverveInput
from csverve.errors import CsverveConcatException
from csverve.errors import CsverveMergeCommonColException
from csverve.errors import CsverveMergeDtypesEmptyMergeSet
from csverve.errors import CsverveMergeException
from csverve.errors import CsverveWriterError
from csverve.errors import DtypesMergeException

from . import test_helpers as helpers


###############################################
#                fixtures                     #
################################################


@pytest.fixture
def n_dtypes():
    return random.randint(2, 10)


@pytest.fixture
def n_frames():
    return random.randint(3, 10)


@pytest.fixture
def n_rows():
    return random.randint(3, 10)


################################################
#                  tests                       #
################################################


class TestAnnotateCsv(helpers.AnnotationHelpers):
    """
    class to test annotate_csv
    """

    def test_annotate_csv(self, tmpdir, n_rows):
        """
        basic sanity check - test annotating normal csv
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """
        dtypes = {v: "int" for v in 'ABCD'}
        dtypes["cell_id"] = "str"
        ann_dtypes = {v: "int" for v in 'ERF'}

        csv, annotation, annotated = self.base_annotation_test(tmpdir, n_rows,
                                                               dtypes, ann_dtypes)

        assert self.validate_annotation_test(csv, annotation, annotated, "cell_id")

    def test_annotate_csv_annotate_not_on_cell_id(self, tmpdir, n_rows):
        """
        test annotating on different column that default "cell_id"
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """

        dtypes = {v: "int" for v in 'ABCD'}
        dtypes["TEST"] = "str"
        ann_dtypes = {v: "int" for v in 'ERF'}

        csv, annotation, annotated = self.base_annotation_test(tmpdir, n_rows,
                                                               dtypes, ann_dtypes,
                                                               on="TEST")

        assert self.validate_annotation_test(csv, annotation, annotated, "TEST")

    def test_annotate_csv_annotation_col_mismatch(self, tmpdir, n_rows):
        """
        test annotating csv where annotation_data differs in length from csv
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """

        dtypes = {v: "int" for v in 'ABCD'}
        dtypes["cell_id"] = "str"
        ann_dtypes = {v: "int" for v in 'ERF'}
        annotated = os.path.join(tmpdir, "annotated.csv.gz")

        csv, annotation = self.make_ann_test_inputs(tmpdir, n_rows, dtypes,
                                                    ann_dtypes)

        newdf = pd.Series({"E": 1, "R": 43, "F": 2, "name": "new_cell"}).to_frame().T
        newdf = newdf.set_index('name')

        annotation = pd.concat([annotation, newdf])

        api.annotate_csv(csv, annotation, annotated, ann_dtypes)

        self.validate_annotation_test(csv, annotation, annotated, "cell_id")

    def test_annotate_csv_annotation_col_dtype_mismatch(self, tmpdir, n_rows):
        """
        test annotating csv with inappropriate annotation_dtypes
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """

        dtypes = {v: "int" for v in 'ABCD'}
        dtypes["cell_id"] = "str"
        ann_dtypes = {v: "int" for v in 'ERF'}
        annotated = os.path.join(tmpdir, "annotated.csv.gz")

        csv, annotation = self.make_ann_test_inputs(tmpdir, n_rows, dtypes,
                                                    ann_dtypes)

        api.annotate_csv(csv, annotation, annotated, ann_dtypes)

        csv = CsverveInput(csv).read_csv().merge(annotation)

        assert self.dfs_exact_match(annotated, csv)

    def test_annotate_csv_no_write_header(self, tmpdir, n_rows):
        """
        test annotating csv without writing header
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """
        dtypes = {v: "int" for v in 'ABCD'}
        dtypes["cell_id"] = "str"
        ann_dtypes = {v: "int" for v in 'ERF'}

        csv, annotation, annotated = self.base_annotation_test(tmpdir, n_rows,
                                                               dtypes, ann_dtypes,
                                                               skip_header=True)

        assert self.validate_annotation_test(csv, annotation, annotated, "cell_id")


class TestConcatCsv(helpers.ConcatHelpers):
    """
    test class for csverve concat_csv
    """

    def test_concat_csv(self, tmpdir, n_rows):
        """
        basic sanity check - concat two csvs with same cols
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """
        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, [dtypes, dtypes], write=True,
                                               get_ref=True, dir=tmpdir)

        api.concatenate_csv(csvs, concatenated)

        assert self.dfs_exact_match(ref, concatenated)

    def test_concat_csv_no_header(self, tmpdir, n_rows):
        """
        test concating csvs with no headers
        :param tmpdir: temporary directory to write in
        :param n_rows: number of rows in test csvs
        """

        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, [dtypes, dtypes], write=True,
                                               get_ref=True, dir=tmpdir,
                                               skip_header=True)
        api.concatenate_csv(csvs, concatenated, skip_header=True)

        assert self.dfs_exact_match(ref, concatenated)

        concatenated = pd.read_csv(concatenated)  # ignore separate yaml

        assert all([col not in concatenated.columns.tolist()
                    for col in dtypes.keys()])

    def test_concat_csv_empty_inputs(self, tmpdir):
        """
        test concatenate csv with data-less input csvs
        :param tmpdir: tempdir to test in
        """
        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(0, [dtypes, dtypes], write=True,
                                               get_ref=True, dir=tmpdir)

        api.concatenate_csv(csvs, concatenated)

        assert self.dfs_exact_match(concatenated, ref)

    def test_concat_csv_nothing_to_concat(self, tmpdir):
        """
        test concat_csv with not inputs
        :param tmpdir:  tempdir to test in
        """
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        assert self._raises_correct_error(api.concatenate_csv, [],
                                          concatenated,
                                          expected_error=CsverveConcatException)

    def test_concat_csv_one_file_to_concat(self, tmpdir, n_rows):
        """
        provide just 1 file to concat
        :param tmpdir: temp dir to test in
        :param n_rows: length of test dfs
        :return:
        """
        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        df, csv, ref = self.base_test_concat(n_rows, [dtypes], write=True,
                                             get_ref=True, dir=tmpdir)

        api.concatenate_csv(csv, concatenated)

        assert self.dfs_exact_match(ref, concatenated)

    def test_concat_csv_multiple_files_to_concat(self, tmpdir, n_rows, n_frames):
        """
        provide just 1 file to concat
        """
        dtypes = [{v: "int" for v in 'ABCD'} for _ in range(n_frames)]
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, dtypes, write=True,
                                               get_ref=True, dir=tmpdir)

        api.concatenate_csv(csvs, concatenated)

        assert self.dfs_exact_match(ref, concatenated)

    def test_concat_csv_different_cols(self, tmpdir, n_rows):
        """
        concat two dataframes with different columns
        """
        dtypes1 = {v: "float" for v in 'ABCD'}
        dtypes2 = {v: "float" for v in 'ABGF'}

        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, [dtypes1, dtypes2], write=True,
                                               get_ref=True, dir=tmpdir)

        api.concatenate_csv(csvs, concatenated)

        assert self.dfs_exact_match(ref, concatenated)

    def test_concat_csv_different_dtypes(self, tmpdir, n_rows):
        """
        concat two dataframes same colnames with different dtypes
        """

        dtypes1 = {v: "float" for v in 'ABCD'}
        dtypes2 = {v: "int" for v in 'ABCD'}

        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs = self.base_test_concat(n_rows, [dtypes1, dtypes2], write=True,
                                          get_ref=False, dir=tmpdir)

        assert self._raises_correct_error(api.concatenate_csv, csvs,
                                          concatenated,
                                          expected_error=DtypesMergeException)

    def test_concat_csv_with_nans(self, tmpdir, n_rows):
        """
        concat two csvs with NaNs
        """
        dtypes = {v: "float" for v in 'ABCD'}

        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs = self.make_test_dfs([dtypes, dtypes], n_rows)
        csvs = [os.path.join(tmpdir, "0.csv.gz"),
                os.path.join(tmpdir, "1.csv.gz")]

        dfs[0].iloc[2, dfs[0].columns.get_loc("A")] = np.NaN
        dfs[1].iloc[2, dfs[1].columns.get_loc("A")] = np.NaN

        api.write_dataframe_to_csv_and_yaml(dfs[0], csvs[0], dtypes)
        api.write_dataframe_to_csv_and_yaml(dfs[1], csvs[1], dtypes)

        ref = pd.concat(dfs, ignore_index=True)
        api.concatenate_csv(csvs, concatenated)

        assert self.dfs_exact_match(ref, concatenated)

    def test_concat_csv_gen_nans_with_int_dtype(self, tmpdir, n_rows):
        """
        concat two csvs with NaNs
        """
        dtypes1 = {v: "int" for v in 'ABCD'}
        dtypes2 = {v: "int" for v in 'ABE'}

        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs = self.base_test_concat(n_rows, [dtypes1, dtypes2], write=True,
                                          get_ref=False, dir=tmpdir)

        # check for value error - should raise internal pandas error
        assert self._raises_correct_error(api.concatenate_csv, csvs,
                                          concatenated,
                                          expected_error=CsverveWriterError)


class TestConcatCsvFilesPandas(helpers.ConcatHelpers):
    """
    test class for csverve concat_csv_files_pandas
    """

    def test_concat_csv_pandas(self, tmpdir, n_rows):
        """
        basic sanity check - concat 2 csvs with same cols
        """

        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, [dtypes, dtypes],
                                               write=True, get_ref=True,
                                               dir=tmpdir)

        utils.concatenate_csv_files_pandas(csvs, concatenated, dtypes, True)

        assert self.dfs_exact_match(ref, concatenated)

    def test_concat_csv_pandas_no_header(self, tmpdir, n_rows):
        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, [dtypes, dtypes],
                                               write=True, get_ref=True,
                                               dir=tmpdir, skip_header=True)

        utils.concatenate_csv_files_pandas(csvs, concatenated, dtypes, True)

        assert self.dfs_exact_match(ref, concatenated)


class TestConcatCsvFilesQuickLowMem(helpers.ConcatHelpers):
    """
    test class for csverve concat_csv_files_quick_lowmem
    """

    def test_quick_concat(self, tmpdir, n_rows):
        """
        sanity check - two easy csvs
        """
        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')

        dfs, csvs, ref = self.base_test_concat(n_rows, [dtypes, dtypes],
                                               write=True, get_ref=True,
                                               dir=tmpdir, skip_header=True)

        utils.concatenate_csv_files_quick_lowmem(csvs, concatenated, dtypes, list(dtypes.keys()))

        assert self.dfs_exact_match(ref, concatenated)

    def test_quick_concat_scale(self, tmpdir, n_rows):
        """
        sanity check - two easy csvs
        """
        dtypes = {v: "int" for v in 'ABCD'}
        concatenated = os.path.join(tmpdir, 'concat.csv.gz')
        scale = 100

        dfs, csvs, ref = self.base_test_concat(n_rows * scale, [dtypes, dtypes],
                                               write=True, get_ref=True,
                                               dir=tmpdir, skip_header=True)

        utils.concatenate_csv_files_quick_lowmem(csvs, concatenated, dtypes, list(dtypes.keys()))

        assert self.dfs_exact_match(ref, concatenated)


class TestWriteDataFrameToCsvAndYaml(helpers.WriteHelpers):
    """
    class to test writing of dfs to csv and yamls
    """

    def validate_write_to_csv_yaml_test(self, df, dtypes, filename,
                                        yaml_filename, skipped_header=False):
        """
        tests that write_data_frane_to_csv_and_yaml wrote data correctly
        :param df: input df
        :param dtypes: input dtypes of df
        :param filename: csv file written to
        :param yaml_filename: yaml file written to
        :param wrote_header: T/F wrote header out
        """
        assert os.path.exists(filename)
        assert os.path.exists(yaml_filename)
        assert self.write_file_successful(df, filename, skipped_header=skipped_header)
        assert self.metadata_write_successful(dtypes, yaml_filename)

    def base_write_to_csv_yaml_test(self, temp, dtypes, length, skip_header=False):
        """
        base test for write csv yaml
        """
        df = self.make_test_dfs([dtypes], length)

        csv = self.write_dfs(temp, df, [dtypes], skip_header)

        filename = csv[0]

        yaml_filename = filename + ".yaml"
        os.remove(yaml_filename)

        assert not os.path.exists(yaml_filename)

        api.write_dataframe_to_csv_and_yaml(df[0], filename, dtypes,
                                            skip_header=skip_header)

        return df[0], filename, yaml_filename

    def test_write_to_csv_yaml(self, tmpdir, n_rows):
        """
        basic sanity check - write normal df
        """
        skip_header = False
        dtypes = {v: "int" for v in "ABC"}
        df, filename, yaml_filename = self.base_write_to_csv_yaml_test(
            tmpdir,
            dtypes,
            n_rows,
            skip_header=skip_header
        )

        self.validate_write_to_csv_yaml_test(df, dtypes, filename, yaml_filename,
                                             skipped_header=skip_header)

    def test_write_to_csv_yaml_no_header(self, tmpdir, n_rows):
        """
        write single df without header
        """
        skip_header = False
        dtypes = {v: "int" for v in "ABC"}
        df, filename, yaml_filename = self.base_write_to_csv_yaml_test(
            tmpdir,
            dtypes,
            n_rows,
            skip_header=skip_header)

        self.validate_write_to_csv_yaml_test(
            df, dtypes, filename, yaml_filename,
            skipped_header=skip_header)

    def test_write_to_csv_yaml_empty(self, tmpdir):
        """
        write empty df
        """

        dtypes = {v: "int" for v in 'ABCD'}
        df = pd.DataFrame()
        filename = os.path.join(tmpdir, "df.csv.gz")
        yaml_filename = filename + ".yaml"

        api.write_dataframe_to_csv_and_yaml(df, filename, dtypes)

        assert os.path.exists(filename)
        assert os.path.exists(yaml_filename)


class TestMergeFrames(helpers.MergeHelpers):
    """
    test merge_frames
    """

    def merge_frames_directional_test(self, length, direction):
        """
        merge frames in a given direction; corresponds to "how"
        :param length: length of test dfs
        :param direction: direction to merge in (outter, inner etc.)
        """
        how = direction
        on = ["A"]
        suffs = ["", ""]
        dtypes1 = {v: "int" for v in "ABC"}
        dtypes2 = {v: "int" for v in "ADF"}
        dtypes = [dtypes1, dtypes2]

        dfs, ref = self.base_merge_test(length, how, on, suffs, dtypes)

        merged = utils.merge_frames(dfs, how=how, on=on)

        assert self.dfs_exact_match(ref, merged)

    # TODO: missing case where NaN generated with int column
    def test_merge_frames(self, n_rows):
        """
        basic sanity check - test merging of two frames on 1 col
        """
        self.merge_frames_directional_test(n_rows, "outer")

    def test_merge_frames_inner(self, n_rows):
        """
        test merging of 2 dfs on 1 col with inner merge
        """
        self.merge_frames_directional_test(n_rows, "inner")

    def test_merge_frames_left(self, n_rows):
        """
        test merging of 2 dfs on 1 col with left merge
        """

        self.merge_frames_directional_test(n_rows, "left")

    def test_merge_frames_right(self, n_rows):
        """
        test merging of 2 dfs on 1 col with right merge
        """

        self.merge_frames_directional_test(n_rows, "right")

    def test_merge_frames_multiple_cols(self, n_rows):
        """
        test merging of 2 dfs on multiple columns with right merge
        """

        how = "inner"
        on = ["A", "B"]
        suffs = ["", ""]
        dtypes1 = {v: "int" for v in "ABC"}
        dtypes2 = {v: "int" for v in "ABDF"}
        dtypes = [dtypes1, dtypes2]

        dfs, ref = self.base_merge_test(n_rows, how, on, suffs, dtypes)

        merged = utils.merge_frames(dfs, how=how, on=on)

        assert self.dfs_exact_match(ref, merged)

    def test_merge_frames_one_frame(self, n_rows):
        """
        provide just one df
        :param n_rows: number of rows in simulated df
        :return: assertion
        """

        how = "inner"
        on = ["A"]
        suffs = ["", ""]
        dtypes = {v: "int" for v in "ABC"}

        df = self.base_merge_test(n_rows, how, on, suffs, [dtypes],
                                  get_ref=False)

        merged = utils.merge_frames(df, how=how, on=on)

        assert self.dfs_exact_match(df[0], merged)

    def test_merge_frames_nothing_to_merge(self, n_rows):
        """
        test merging of 2 dfs on 0 col
        """
        dtypes1 = {v: "int" for v in 'CD'}
        dtypes2 = {v: "int" for v in 'EFGH'}

        dfs = self.make_mergeable_test_dfs([dtypes1, dtypes2], [], n_rows)

        assert self._raises_correct_error(utils.merge_frames, dfs,
                                          how="outer", on=[],
                                          expected_error=CsverveMergeException)

    def test_merge_frames_cols_to_merge_have_different_dtypes(self, n_rows):
        """
        test merging of 2 dfs on 1 col. Each dataframe has different dtype for col
        """

        dtypes1 = {v: "int" for v in 'ABCD'}
        dtypes2 = {v: "str" for v in 'AEFGH'}

        dfs = self.make_mergeable_test_dfs([dtypes1, dtypes2], [], n_rows)

        assert self._raises_correct_error(utils.merge_frames, dfs,
                                          how="outer", on=[],
                                          expected_error=CsverveMergeException)

    def test_merge_frames_no_merge_col_given(self, n_rows):
        """
        test merging of 2 dfs on no col
        """

        dtypes1 = {v: "int" for v in 'ACD'}
        dtypes2 = {v: "int" for v in 'ABG'}

        dfs = self.make_mergeable_test_dfs([dtypes1, dtypes2], [], n_rows)

        assert self._raises_correct_error(utils.merge_frames, dfs,
                                          how="outer", on=[],
                                          expected_error=CsverveMergeException)

    def test_merge_frames_with_nans(self, n_rows):
        """
        test merging of 2 dfs on 1 col which contains NaNs in each
        """
        dtypes1 = {v: "float" for v in 'ACD'}
        dtypes2 = {v: "float" for v in 'AEG'}
        how = "outer"
        on = ['A']

        dfs = self.make_mergeable_test_dfs([dtypes1, dtypes2], on, n_rows)

        dfs[0].iloc[2, dfs[0].columns.get_loc(on[0])] = np.NaN
        dfs[1].iloc[2, dfs[1].columns.get_loc(on[0])] = np.NaN

        ref = dfs[0].merge(dfs[1], how=how, on=on)

        merged = utils.merge_frames(dfs, how=how, on=on)

        assert self.dfs_exact_match(ref, merged)

    def test_merge_frames_differing_vals_on_common_cols(self, n_rows):
        """
        test merging of 2 dfs on multiple columns with right merge
        """
        how = "inner"
        on = ["A"]
        dtypes1 = {v: "float" for v in "AC"}
        dtypes2 = {v: "float" for v in "ACF"}

        dfs = self.make_mergeable_test_dfs([dtypes1, dtypes2], on, n_rows)

        assert self._raises_correct_error(utils.merge_frames, dfs,
                                          how=how, on=on,
                                          expected_error=CsverveMergeCommonColException)


class TestMergeDtypes(helpers.MergeHelpers):

    def test_merge_dtypes(self):
        """
        basic sanity check - test merging of two dtype dicts
        """
        dtypes1 = {v: "float" for v in 'ACD'}
        dtypes2 = {v: "float" for v in 'ACDEF'}
        ref = {v: "float" for v in
               set(dtypes1.keys()).union(set(dtypes2.keys()))}

        merged_dtypes = utils.merge_dtypes([dtypes1, dtypes2])

        assert ref == merged_dtypes

    def test_merge_dtypes_none_given(self):
        """
        test merging of empty list of dtypes
        """
        assert self._raises_correct_error(utils.merge_dtypes, [],
                                          expected_error=CsverveMergeDtypesEmptyMergeSet)

    def test_merge_dtypes_one_given(self):
        """
        test merging of list of 1 dtype dict
        """
        dtypes1 = {v: "float" for v in 'ACD'}
        ref = dtypes1

        merged_dtypes = utils.merge_dtypes([dtypes1])

        assert ref == merged_dtypes

    def test_merge_dtypes_multiple_given(self, n_dtypes):
        """
        test merging of n_dtypes dtype dicts
        :param n_dtypes: number of dtypes to merge
        """
        dtypes = [{v: "int" for v in "".join(self._str_list(3, "A"))}
                  for _ in range(n_dtypes)]

        # taken from https://stackoverflow.com/questions/9819602/union-of-dict-objects-in-python
        ref = dict(itertools.chain.from_iterable(dct.items()
                                                 for dct in dtypes))

        merged_dtypes = utils.merge_dtypes(dtypes)

        assert ref == merged_dtypes

    @staticmethod
    def merge_dtype_test_types_different(dtype):
        """
        test merge dtypes that are different
        """
        dtypes1 = {v: dtype for v in 'ACD'}
        dtypes2 = {v: dtype for v in 'ACDEF'}
        ref = {v: dtype for v in
               set(dtypes1.keys()).union(set(dtypes2.keys()))}

        merged_dtypes = utils.merge_dtypes([dtypes1, dtypes2])

        assert ref == merged_dtypes

    def test_merge_dtypes_float(self):
        """
        merge dtypes with float types
        """
        self.merge_dtype_test_types_different("float")

    def test_merge_dtypes_str(self):
        """
        merge dtypes with str types
        """
        self.merge_dtype_test_types_different("str")

    def test_merge_dtypes_bool(self):
        """
        merge dtypes with bool types
        """
        self.merge_dtype_test_types_different("bool")

    @staticmethod
    def merge_dtype_test_types_mixed():
        """
        test merge dtypes that have mixed typing
        """
        types = ["int", "float", "bool", "str"]
        dtypes1 = {v: random.choice(types) for v in "ACD"}
        dtypes2 = {v: random.choice(types) for v in "ACD"}

        ref = {v: type for v in
               set(dtypes1.keys()).union(set(dtypes2.keys()))}

        merged_dtypes = utils.merge_dtypes([dtypes1, dtypes2])

        assert ref == merged_dtypes


class TestMergeCsv(helpers.MergeHelpers):
    """
    class to test merge_csv
    """

    def test_merge_csv(self, tmpdir, n_rows):
        """
        test merge_csv
        """
        dtypes1 = {v: "int" for v in 'ABCD'}
        dtypes2 = {v: "int" for v in 'AEFGH'}
        how = "outer"
        on = ["A"]
        suffs = ["", ""]
        merged = os.path.join(tmpdir, "merged.csv.gz")

        dfs, csvs, ref = self.base_merge_test(n_rows, how, on, suffs,
                                              [dtypes1, dtypes2],
                                              write=True, dir=tmpdir)

        api.merge_csv(csvs, merged, how=how, on=on)

        assert os.path.exists(merged)

        assert self.dfs_exact_match(ref, merged)

    def test_merge_csv_no_header(self, tmpdir, n_rows):
        """
        merge csvs with no headers
        """

        dtypes1 = {v: "int" for v in 'ABCD'}
        dtypes2 = {v: "int" for v in 'AEFGH'}
        how = "outer"
        on = ["A"]
        suffs = ["", ""]
        merged = os.path.join(tmpdir, "merged.csv.gz")

        dfs, csvs, ref = self.base_merge_test(n_rows, how, on, suffs,
                                              [dtypes1, dtypes2],
                                              write=True, dir=tmpdir,
                                              skip_header=True)

        api.merge_csv(csvs, merged, how=how, on=on)

        assert os.path.exists(merged)

        assert self.dfs_exact_match(ref, merged)


class TestCsverve(unittest.TestCase):
    """Tests for `csverve` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""

    def test_command_line_interface(self):
        """Test the CLI."""
        runner = CliRunner()
        _ = runner.invoke(cli.cli)
        # assert result.exit_code == 0
        # assert 'csverve.cli.main' in result.output
        # help_result = runner.invoke(cli.main, ['--help'])
        # assert help_result.exit_code == 0
        # assert '--help  Show this message and exit.' in help_result.output



import os
from typing import List, Dict, Union

import csverve.utils as utils
import pandas as pd  # type: ignore
from csverve.core import CsverveInput
from csverve.core import CsverveOutputDataFrame
from csverve.core import CsverveOutputFileStream
from csverve.core import IrregularCsverveInput
from csverve.errors import CsverveConcatException
from csverve.utils import concatenate_csv_files_quick_lowmem, concatenate_csv_files_pandas

from warnings import warn

def get_columns(infile):
    return CsverveInput(infile).columns


def get_dtypes(infile):
    return CsverveInput(infile).dtypes


def rewrite_csv_file(
        filepath: str,
        outputfile: str,
        skip_header: bool = False,
        dtypes: Dict[str, str] = None,
        **kwargs
) -> None:
    """
    Generate header less csv files.

    @param filepath: File path of CSV.
    @param outputfile: File path of header less CSV to be generated.
    @param skip_header: boolean, True = write header, False = don't write header.
    @param dtypes: Dictionary of pandas dtypes, where key = column name, value = dtype.
    @return:
    """
    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    if os.path.exists(filepath + '.yaml'):
        csvinput: Union[CsverveInput, IrregularCsverveInput] = CsverveInput(filepath)
        df = csvinput.read_csv()

        csvoutput_df = CsverveOutputDataFrame(
            df, outputfile, skip_header=skip_header,
            dtypes=csvinput.dtypes
        )
        csvoutput_df.write_df()
    else:
        assert dtypes
        csvinput = IrregularCsverveInput(filepath, dtypes)

        csvoutput_fs = CsverveOutputFileStream(
            outputfile, skip_header=skip_header, columns=csvinput.columns,
            dtypes=csvinput.dtypes
        )
        csvoutput_fs.rewrite_csv(filepath)


def merge_csv(
        in_filenames: Union[List[str], Dict[str, str]],
        out_filename: str,
        how: str,
        on: List[str],
        skip_header: bool = False,
        **kwargs
) -> None:
    """
    Create one gzipped CSV out of multiple gzipped CSVs.

    @param in_filenames: Dictionary containing file paths as keys
    @param out_filename: Path to newly merged CSV
    @param how: How to join DataFrames (inner, outer, left, right).
    @param on: Column(s) to join on, comma separated if multiple.
    @param skip_header: boolean, True = write header, False = don't write header
    @return:
    """
    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    if isinstance(in_filenames, dict):
        in_filenames = list(in_filenames.values())

    data: List[CsverveInput] = [CsverveInput(infile) for infile in in_filenames]

    dfs: List[str] = [csvinput.read_csv() for csvinput in data]

    dtypes: List[Dict[str, str]] = [csvinput.dtypes for csvinput in data]

    merged_data: pd.DataFrame = utils.merge_frames(dfs, how, on)

    dtypes_: Dict[str, str] = utils.merge_dtypes(dtypes)

    csvoutput: CsverveOutputDataFrame = CsverveOutputDataFrame(
        merged_data, out_filename, dtypes_, skip_header=skip_header
    )
    csvoutput.write_df()


def concatenate_csv(inputfiles: List[str], output: str, skip_header: bool = False,
                    drop_duplicates: bool = False, **kwargs) -> None:
    """
    Concatenate gzipped CSV files, dtypes in meta YAML files must be the same.

    @param inputfiles: List of gzipped CSV file paths, or a dictionary where the keys are file paths.
    @param output: Path of resulting concatenated gzipped CSV file and meta YAML.
    @param skip_header: boolean, True = write header, False = don't write header.
    @return:
    """
    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    if isinstance(inputfiles, dict):
        inputfiles = list(inputfiles.values())

    if inputfiles == []:
        raise CsverveConcatException("nothing provided to concat")

    inputs: List[CsverveInput] = [CsverveInput(infile) for infile in inputfiles]

    dtypes: Dict[str, str] = utils.merge_dtypes([csvinput.dtypes for csvinput in inputs])

    headers: List[bool] = [csvinput.header for csvinput in inputs]

    columns: List[List[str]] = [csvinput.columns for csvinput in inputs]

    low_memory: bool = True
    if any(headers):
        low_memory = False

    if not all(columns[0] == elem for elem in columns):
        low_memory = False

    if drop_duplicates:
        low_memory = False

    if low_memory:
        concatenate_csv_files_quick_lowmem(inputfiles, output, dtypes, columns[0], skip_header=skip_header)
    else:
        concatenate_csv_files_pandas(inputfiles, output, dtypes, skip_header=skip_header,
                                     drop_duplicates=drop_duplicates)


def annotate_csv(
        infile: str,
        annotation_df: pd.DataFrame,
        outfile,
        annotation_dtypes,
        on="cell_id",
        skip_header: bool = False,
        **kwargs
):
    """
    TODO: fill this in
    @param infile:
    @param annotation_df:
    @param outfile:
    @param annotation_dtypes:
    @param on:
    @param skip_header:
    @return:
    """

    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    csvinput = CsverveInput(infile)
    metrics_df = csvinput.read_csv()

    # get annotation rows that correspond to rows in on
    reformed_annotation = annotation_df[annotation_df[on].isin(metrics_df[on])]

    # do nothing if the annotation df is empty
    if reformed_annotation.empty:  # so we dont add NaNs
        return write_dataframe_to_csv_and_yaml(metrics_df, outfile,
                                               csvinput.dtypes,
                                               skip_header=skip_header)

    metrics_df = metrics_df.merge(reformed_annotation, on=on, how='outer')

    csv_dtypes = csvinput.dtypes

    for col, dtype in csv_dtypes.items():
        if col in annotation_dtypes:
            assert dtype == annotation_dtypes[col]

    csv_dtypes.update(annotation_dtypes)

    output = CsverveOutputDataFrame(metrics_df, outfile, csv_dtypes, skip_header=skip_header)
    output.write_df()


def simple_annotate_csv(
        in_f: str,
        out_f: str,
        col_name: str,
        col_val: str,
        col_dtype: str,
        skip_header: bool = False,
        **kwargs
) -> None:
    """
    Simplified version of the annotate_csv method.
    Add column with the same value for all rows.

    @param in_f:
    @param out_f:
    @param col_name:
    @param col_val:
    @param col_dtype:
    @param skip_header:
    @return:
    """
    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    csvinput = CsverveInput(in_f)
    metrics_df = csvinput.read_csv()
    metrics_df[col_name] = col_val

    csv_dtypes = csvinput.dtypes
    csv_dtypes[col_name] = col_dtype

    output = CsverveOutputDataFrame(metrics_df, out_f, csv_dtypes, skip_header=skip_header)
    output.write_df()


def add_col_from_dict(
        infile,
        col_data,
        outfile,
        dtypes,
        skip_header=False,
        **kwargs
):
    """
    TODO: fill this in
    Add column to gzipped CSV.

    @param infile:
    @param col_data:
    @param outfile:
    @param dtypes:
    @param skip_header:
    @return:
    """

    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    csvinput = CsverveInput(infile)
    csv_dtypes = csvinput.dtypes
    csvinput = csvinput.read_csv()

    for col_name, col_value in col_data.items():
        csvinput[col_name] = col_value

    dtypes = utils.merge_dtypes([csv_dtypes, dtypes])
    output = CsverveOutputDataFrame(
        csvinput, outfile, dtypes, skip_header=skip_header
    )
    output.write_df()


def write_dataframe_to_csv_and_yaml(
        df: pd.DataFrame,
        outfile: str,
        dtypes: Dict[str, str],
        skip_header: bool = False,
        **kwargs
) -> None:
    """
    Output pandas dataframe to a CSV and meta YAML files.

    @param df: pandas DataFrame.
    @param outfile: Path of CSV & YAML file to be written to.
    @param dtypes: dictionary of pandas dtypes by column, keys = column name, value = dtype.
    @param skip_header: boolean, True = skip writing header, False = write header
    @return:
    """

    if kwargs.get('write_header') is not None:
        warn('write_header has been deprecated and will be ignored, please use skip_header instead', DeprecationWarning)

    csvoutput: CsverveOutputDataFrame = CsverveOutputDataFrame(
        df, outfile, dtypes, skip_header=skip_header
    )
    csvoutput.write_df()


def read_csv(infile: str, chunksize: int = None, usecols=None, dtype=None) -> pd.DataFrame:
    """
    Read in CSV file and return as a pandas DataFrame.

    Assumes a YAML meta file in the same path with the same name, with a .yaml extension.
    YAML file structure is atop this file.

    @param infile: Path to CSV file.
    @param chunksize: Number of rows to read at a time (optional, applies to large datasets).
    @param usecols: Restrict to specific columns (optional).
    @param dtype: Override the dtypes on specific columns (optional).
    @return: pandas DataFrame.
    """
    return CsverveInput(infile).read_csv(chunksize=chunksize, usecols=usecols, dtype=dtype)



from csverve.utils.utils import pandas_to_std_types
from csverve.utils.utils import merge_dtypes
from csverve.utils.utils import concatenate_csv_files_pandas
from csverve.utils.utils import concatenate_csv_files_quick_lowmem
from csverve.utils.utils import merge_frames



from csverve.api.api import rewrite_csv_file
from csverve.api.api import merge_csv
from csverve.api.api import concatenate_csv
from csverve.api.api import annotate_csv
from csverve.api.api import simple_annotate_csv
from csverve.api.api import add_col_from_dict
from csverve.api.api import write_dataframe_to_csv_and_yaml
from csverve.api.api import read_csv
from csverve.api.api import get_columns
from csverve.api.api import get_dtypes



from csverve.errors.errors import CsverveMergeDtypesEmptyMergeSet
from csverve.errors.errors import CsverveMergeCommonColException
from csverve.errors.errors import DtypesMergeException
from csverve.errors.errors import CsverveConcatException
from csverve.errors.errors import CsverveAnnotateError
from csverve.errors.errors import CsverveMergeException
from csverve.errors.errors import CsverveMergeColumnMismatchException
from csverve.errors.errors import CsverveParseError
from csverve.errors.errors import CsverveInputError
from csverve.errors.errors import CsverveWriterError
from csverve.errors.errors import CsverveDtypeError



from typing import List, Dict

import pandas as pd  # type: ignore
from csverve.core import CsverveOutput
from csverve.errors import CsverveWriterError


class CsverveOutputDataFrame(CsverveOutput):
    def __init__(
            self,
            df: pd.DataFrame,
            filepath: str,
            dtypes: Dict[str, str],
            skip_header: bool = False,
            na_rep: str = 'NaN',
            sep: str = ',',
    ) -> None:
        """
        CSV file and all related metadata.

        @param filepath: CSV file path.
        @param dtypes: Dictionary of pandas dtypes, where key = column name, value = dtype.
        @param header: boolean, True = write header, False = don't write header.
        @param na_rep: replace NaN with this value.
        @param columns: List of column names.
        """
        self.df = df
        columns: List[str] = list(self.df.columns.values)

        super().__init__(
            filepath, dtypes, columns,
            skip_header=skip_header, na_rep=na_rep, sep=sep
        )

        self._cast_df()

    def _cast_df(self):
        """
        Cast dataframe dtypes.

        @return:
        """

        for column_name in self.df.columns.values:
            try:
                self.df[column_name] = self.df[column_name].astype(self.dtypes[column_name])
            except ValueError:
                raise CsverveWriterError(
                    'Unable to cast the column {} to {}'.format(column_name, self.dtypes[column_name])
                )

    def write_df(self) -> None:
        """
        Write out dataframe to CSV.

        @param df: Pandas DataFrames.
        @param chunks: bool.
        @return:
        """

        if isinstance(self.df, pd.io.parsers.TextFileReader):
            for i, data in enumerate(self.df):
                if i == 0:
                    data.to_csv(
                        self.filepath, sep=self.sep, na_rep=self.na_rep,
                        index=False, compression='gzip', header=(not self.skip_header), mode='w'
                    )
                else:
                    data.to_csv(
                        self.filepath, sep=self.sep, na_rep=self.na_rep,
                        index=False, compression='gzip', header=False, mode='a'
                    )
        elif isinstance(self.df, pd.DataFrame):
            self.df.to_csv(
                self.filepath, sep=self.sep, na_rep=self.na_rep,
                index=False, compression='gzip', header=(not self.skip_header)
            )
        else:
            raise CsverveWriterError('Invalid df provided as input')

        self.write_yaml()



#!/usr/bin/env python

"""The setup script."""

import versioneer

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'click>=7.0',
    'pandas>=1.0.0',
    'pyyaml',
    'numpy',
    'pytest',
    'Sphinx==1.8.5',
    'mypy==0.790',
]

setup_requirements = []

test_requirements = []

setup(
    author="Shah Lab",
    author_email='todo@todo.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="todo: add description",
    entry_points={
        'console_scripts': [
            'csverve=csverve.cli:cli',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='csverve',
    name='csverve',
    packages=find_packages(include=['csverve', 'csverve.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mondrian-scwgs/csverve',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
)
